import importlib


def is_neptune_available():
    return importlib.util.find_spec("neptune") is not None


def is_layer_available():
    return importlib.util.find_spec("layer") is not None


def check_requirements(package_names):
    """
    Raise error if module is not installed.
    """
    missing_packages = []
    for package_name in package_names:
        if importlib.util.find_spec(package_name) is None:
            missing_packages.append(package_name)
    if missing_packages:
        raise ImportError(f"The following packages are required to use this module: {missing_packages}")
    yield



from video_transformers.auto.backbone import AutoBackbone
from video_transformers.auto.head import AutoHead
from video_transformers.auto.neck import AutoNeck
from video_transformers.modules import TimeDistributed, VideoClassificationModel

__version__ = "0.0.4"



from pathlib import Path
from typing import Dict, Union

import numpy as np
import torch
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.tracking import GeneralTracker
from torch.optim.lr_scheduler import _LRScheduler
from tqdm.auto import tqdm

import video_transformers.data
from video_transformers.modules import VideoClassificationModel
from video_transformers.schedulers import get_linear_scheduler_with_warmup
from video_transformers.tasks.single_label_classification import SingleLabelClassificationTaskMixin
from video_transformers.tracking import TensorBoardTracker
from video_transformers.utils.extra import scheduler_to_config
from video_transformers.utils.file import increment_path

logger = get_logger(__name__)


class BaseTrainer:
    def __init__(
        self,
        datamodule: video_transformers.data.VideoDataModule,
        model: VideoClassificationModel,
        max_epochs: int = 12,
        cpu: bool = False,
        mixed_precision: str = "no",
        output_dir: str = "runs",
        seed: int = 42,
        trackers: list[GeneralTracker] = None,
        checkpoint_save: bool = True,
        checkpoint_save_interval: int = 1,
        checkpoint_save_policy: str = "epoch",
        experiment_name="exp",
        optimizer: torch.optim.Optimizer = None,
        scheduler: _LRScheduler = None,
        config_dict: dict = None,
        loss_function: torch.nn.modules.loss._Loss = None,
    ):
        """
        Args:
            datamodule: A `VideoDataModule` instance.
            model: A `nn.Module` instance.
            max_epochs:
            cpu: Use CPU if True.
            mixed_precision: One of ['no', 'fp16', 'bf16'].
            output_dir: The directory to save outputs.
            lr: The max learning rate.
            num_epochs: The number of epochs.
            seed: The random seed.
            trackers: A list of `GeneralTracker` instances.
            checkpoint_save: Whether to save checkpoints.
            checkpoint_save_interval: The interval between saving checkpoints.
            checkpoint_save_policy: One of ['epoch', 'step'].
            experiment_name: The name of the experiment.
            optimizer: The optimizer.
            scheduler: The scheduler.
            config_dict: Optional config setting to be logged.
        """
        self.experiment_dir = increment_path(Path(output_dir) / experiment_name, exist_ok=False)

        self._set_optimizer(model, optimizer)
        self._set_scheduler(scheduler, max_epochs)
        self._set_loss_function(loss_function)
        self._set_trackers(trackers, experiment_name)

        self.accelerator = Accelerator(
            cpu=cpu, mixed_precision=mixed_precision, log_with=self.trackers, logging_dir=self.experiment_dir
        )

        labels = datamodule.labels

        # set haparams to be logged
        ignore_args = [
            "self",
            "ignore_args",
            "model",
            "datamodule",
            "output_dir",
            "experiment_name",
            "optimizer",
            "scheduler",
            "config_dict",
            "loss_function",
            "trackers",
            "vt_version",
        ]
        vt_version = {"video_transformers_version": video_transformers.__version__}
        hparams = locals()
        hparams = {k: v for k, v in hparams.items() if k not in ignore_args}
        if config_dict is not None:
            hparams.update(config_dict)
        if isinstance(model, VideoClassificationModel):
            hparams.update({"model": model.config})
        if isinstance(datamodule, video_transformers.data.VideoDataModule):
            hparams.update({"data": datamodule.config})
        if isinstance(self.scheduler, torch.optim.lr_scheduler.SequentialLR):
            hparams.update({"scheduler": scheduler_to_config(self.scheduler)})
        hparams.update(vt_version)
        self.hparams = hparams

        # We need to initialize the trackers we use, and also store our configuration
        if self.accelerator.is_main_process:
            self.accelerator.init_trackers(experiment_name, hparams)

        # Set the seed before splitting the data.
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

        (
            self.model,
            self.optimizer,
            self.train_dataloader,
            self.val_dataloader,
            self.test_dataloader,
            self.scheduler,
        ) = self.accelerator.prepare(
            model,
            self.optimizer,
            datamodule.train_dataloader,
            datamodule.val_dataloader,
            datamodule.test_dataloader,
            self.scheduler,
        )

        self.overall_step = 0
        self.overall_epoch = 0
        self.starting_epoch = 0

        self.labels = labels

        self.last_saved_checkpoint_epoch = -1

    @property
    def train_metrics(self):
        raise NotImplementedError()

    @property
    def val_metrics(self):
        raise NotImplementedError()

    @property
    def last_train_result(self):
        return None

    @property
    def last_val_result(self):
        return None

    def _set_optimizer(self, model, optimizer):
        if optimizer is None:
            optimizer = torch.optim.AdamW(model.parameters(), lr=1e-4)
        self.optimizer = optimizer

    def _set_scheduler(self, scheduler, max_epochs=None):
        if scheduler is None:
            scheduler = get_linear_scheduler_with_warmup(self.optimizer, max_epochs=max_epochs)
        self.scheduler = scheduler

    def _set_loss_function(self, loss_function):
        if loss_function is None:
            loss_function = torch.nn.CrossEntropyLoss()
        self.loss_function = loss_function

    def _set_trackers(self, trackers, experiment_name: str):
        tensorboard_present = False
        tb_tracker = TensorBoardTracker(run_name=experiment_name, logging_dir=self.experiment_dir)
        if trackers is None:
            trackers = tb_tracker
            tensorboard_present = True
        else:
            for tracker in trackers:
                if tracker.name == "tensorbaord":
                    tracker = tb_tracker
                    tensorboard_present = True
        if not tensorboard_present:
            trackers.append(tb_tracker)
        self.trackers = trackers

    def _log_last_lr(self):
        if self.accelerator.is_main_process:
            lr = self.optimizer.param_groups[-1]["lr"]
            self.accelerator.log({"lr": lr}, step=self.overall_epoch)

    def _get_last_train_score(self) -> Union[None, Dict[str, float]]:
        if self.last_train_result is not None:
            first_train_metric = list(self.last_train_result.keys())[0]
            return ("train/" + first_train_metric, self.last_train_result[first_train_metric].mean())
        else:
            return None

    def _get_last_val_score(self) -> Union[None, Dict[str, float]]:
        if self.last_val_result is not None:
            first_val_metric = list(self.last_val_result.keys())[0]
            return ("val/" + first_val_metric, self.last_val_result[first_val_metric].mean())
        else:
            return None

    def _save_state_and_checkpoint(self):
        if self.accelerator.is_main_process:
            output_name = "checkpoint"
            save_path = Path(self.experiment_dir) / output_name
            if self.hparams["checkpoint_save_policy"] in ["steps", "step"]:
                if self.overall_step % self.hparams["checkpoint_save_interval"] == 0:
                    self.accelerator.save_state(save_path)
            elif self.hparams["checkpoint_save_policy"] in ["epochs", "epoch"]:
                if (
                    self.last_saved_checkpoint_epoch != self.overall_epoch
                    and self.overall_epoch % self.hparams["checkpoint_save_interval"] == 0
                ):
                    self.accelerator.save_state(save_path)
                    self.save_checkpoint(save_path)
                    self.last_saved_checkpoint_epoch = self.overall_epoch
            else:
                raise ValueError(f"Unknown checkpoint save policy: {self.hparams['checkpoint_save_policy']}")

    def save_checkpoint(self, save_path: Union[str, Path]):
        config = self.model.config.copy()
        data_config = self.hparams["data"]
        config.update(
            {
                "preprocess_means": data_config["preprocess_config"]["means"],
                "preprocess_stds": data_config["preprocess_config"]["stds"],
                "preprocess_min_short_side_scale": data_config["preprocess_config"]["min_short_side_scale"],
                "preprocess_input_size": data_config["preprocess_config"]["input_size"],
                "num_timesteps": data_config["num_timesteps"],
                "labels": data_config["labels"],
            }
        )
        self.model.save_pretrained(save_path, config)

    def _update_state(self, loss):
        self.accelerator.backward(loss)
        self.optimizer.step()
        self.optimizer.zero_grad()
        self.overall_step += 1

    def _one_train_loop(self, pbar, len_train_dataloader):
        pbar.set_description(f"Epoch {self.overall_epoch} (Train)")

        self.model.train()
        train_loss = 0
        train_dataloader_iter = iter(self.train_dataloader)

        for step in range(len_train_dataloader):
            pbar.update(1)

            batch = next(train_dataloader_iter)
            loss = self.training_step(batch)
            train_loss += loss

            # update accelerator, optimizer, scheduler
            self._update_state(loss)

            # update pbar with train batch loss
            pbar.set_postfix({"loss": f"{loss:.4f}"})

            self._save_state_and_checkpoint()

        return train_loss

    def _one_val_loop(self, pbar, len_val_dataloader):
        pbar.set_description(f"Epoch {self.overall_epoch} (Val)  ")

        self.model.eval()
        val_loss = 0
        val_dataloader_iter = iter(self.val_dataloader)

        for step in range(len_val_dataloader):
            pbar.update(1)

            batch = next(val_dataloader_iter)
            loss = self.validation_step(batch)
            val_loss += loss

            # update pbar with val batch loss
            train_score = self._get_last_train_score()
            val_score = self._get_last_val_score()
            if val_score is not None:
                pbar.set_postfix(
                    {
                        "loss": f"{loss:.4f}",
                        train_score[0]: f"{train_score[1]:.3f}",
                    }
                )
            else:
                pbar.set_postfix(
                    {
                        "loss": f"{loss:.4f}",
                        train_score[0]: f"{train_score[1]:.3f}",
                    }
                )

        return val_loss

    def fit(self):
        logger.info("Calculating training & validation sizes for better experience...")
        len_train_dataloader = len(self.train_dataloader)
        len_val_dataloader = len(self.val_dataloader)

        for epoch in range(self.starting_epoch, self.hparams["max_epochs"]):
            self._log_last_lr()

            with tqdm(
                total=len_train_dataloader + len_val_dataloader,
                unit=" batch",
                disable=not self.accelerator.is_local_main_process,
            ) as pbar:

                # train loop
                train_loss = self._one_train_loop(pbar, len_train_dataloader)

                # call the end of train epoch hook
                self.on_training_epoch_end()

                # update pbar with train epoch loss and score
                train_loss = train_loss / len_train_dataloader
                train_score = self._get_last_train_score()
                pbar.set_postfix({"loss": f"{train_loss:.4f}", train_score[0]: f"{train_score[1]:.3f}"})

                # val loop
                val_loss = self._one_val_loop(pbar, len_val_dataloader)

                # call the end of val epoch hook
                self.on_validation_epoch_end()

                # update pbar with val epoch loss and score
                val_loss = val_loss / len_val_dataloader
                val_score = self._get_last_val_score()
                pbar.set_postfix(
                    {
                        "loss": f"{val_loss:.4f}",
                        val_score[0]: f"{val_score[1]:.3f}",
                        train_score[0]: f"{train_score[1]:.3f}",
                    }
                )
                pbar.set_description(f"Epoch {self.overall_epoch} (Done) ")

            self._save_state_and_checkpoint()
            self.overall_epoch = int(self.overall_epoch + 1)
            self.scheduler.step()


class SingleLabelClassificationTrainer(SingleLabelClassificationTaskMixin, BaseTrainer):
    pass


def trainer_factory(task: str):
    if task == "single_label_classification":
        return SingleLabelClassificationTrainer
    else:
        raise ValueError(f"Unknown trainer task: {task}")



from typing import Dict

from video_transformers.necks import BaseNeck


class AutoNeck:
    """
    AutoNeck is a class that automatically instantiates a video model neck from a config.

    """

    @classmethod
    def from_config(cls, config: Dict) -> BaseNeck:
        neck_class_name = config.get("name")
        if neck_class_name == "LSTMNeck":
            from video_transformers.necks import LSTMNeck

            num_features = config.get("num_features")
            hidden_size = config.get("hidden_size")
            num_layers = config.get("num_layers")
            return_last = config.get("return_last")
            return LSTMNeck(num_features, hidden_size, num_layers, return_last)
        elif neck_class_name == "GRUNeck":
            from video_transformers.necks import GRUNeck

            num_features = config.get("num_features")
            hidden_size = config.get("hidden_size")
            num_layers = config.get("num_layers")
            return_last = config.get("return_last")
            return GRUNeck(num_features, hidden_size, num_layers, return_last)
        elif neck_class_name == "TransformerNeck":
            from video_transformers.necks import TransformerNeck

            num_features = config.get("num_features")
            num_timesteps = config.get("num_timesteps")
            transformer_enc_num_heads = config.get("transformer_enc_num_heads")
            transformer_enc_num_layers = config.get("transformer_enc_num_layers")
            transformer_enc_act = config.get("transformer_enc_act")
            dropout_p = config.get("dropout_p")
            return_mean = config.get("return_mean")
            return TransformerNeck(
                num_features,
                num_timesteps,
                transformer_enc_num_heads,
                transformer_enc_num_layers,
                transformer_enc_act,
                dropout_p,
                return_mean,
            )

        else:
            raise ValueError(f"Unsupported neck class name: {neck_class_name}")



import os
from typing import Optional, Union

from accelerate.logging import get_logger
from accelerate.tracking import GeneralTracker, is_tensorboard_available

from video_transformers.utils.imports import check_requirements, is_layer_available, is_neptune_available
from video_transformers.utils.logger import _flatten_dict, _sanitize_params

logger = get_logger(__name__)

if is_neptune_available():
    import neptune.new as neptune

if is_layer_available():
    import layer
    import layer.contracts.projects

if is_tensorboard_available():
    from accelerate.tracking import tensorboard

# TODO: log image (confusion mat)


class NeptuneTracker(GeneralTracker):
    """
    A `Tracker` class that supports `neptune`. Should be initialized at the start of your script.
    Args:
        run_name (`str`):
            The name of the experiment run.
        kwargs:
            Additional key word arguments passed along to the `neptune.init` method.
    """

    name = "neptune"
    requires_logging_directory = False

    def __init__(self, run_name: str, **kwargs):
        check_requirements(["neptune-client"])

        self.run_name = run_name

        self.run: neptune.Run = neptune.init(name=self.run_name, **kwargs)
        logger.info(f"Initialized Neptune project {self.run_name}")
        logger.info("Make sure to log any initial configurations with `self.store_init_configuration` before training!")

    @property
    def tracker(self):
        return self.run

    def store_init_configuration(self, values: dict):
        """
        Logs `values` as hyperparameters for the run. Should be run at the beginning of your experiment.
        Args:
            values (Dictionary `str` to `bool`, `str`, `float` or `int`):
                Values to be stored as initial hyperparameters as key-value pairs. The values need to have type `bool`,
                `str`, `float`, `int`, or `None`.
        """
        self.run["config"] = values
        logger.info("Stored initial configuration hyperparameters to Neptune")

    def log(self, values: dict, step: Optional[int] = None, **kwargs):
        """
        Logs `values` to the current run.
        Args:
            values (Dictionary `str` to `str`, `float`, `int` or `dict` of `str` to `float`/`int`):
                Values to be logged as key-value pairs. The values need to have type `str`, `float`, `int` or `dict` of
                `str` to `float`/`int`.
            step (`int`, *optional*):
                The run step. If included, the log will be affiliated with this step.
            kwargs:
                Additional key word arguments passed along to the `neptune.log` method.
        """
        for key, value in values.items():
            if isinstance(value, (dict, int, float, bool, str)):
                self.run[key].log(value, step=step, **kwargs)
            else:
                raise NotImplementedError(f"NeptuneTracker does not support logging {type(value)}")
        logger.info("Successfully logged to Neptune")

    def finish(self):
        """
        Closes `neptune` writer
        """
        self.run.stop()
        logger.info("Neptune run closed")


class TensorBoardTracker(GeneralTracker):
    """
    A `Tracker` class that supports `tensorboard`. Should be initialized at the start of your script.

    Args:
        run_name (`str`):
            The name of the experiment run
        logging_dir (`str`, `os.PathLike`):
            Location for TensorBoard logs to be stored.
    """

    name = "tensorboard"
    requires_logging_directory = True

    def __init__(self, run_name: str, logging_dir: Optional[Union[str, os.PathLike]]):
        self.run_name = run_name
        self.logging_dir = os.path.join(logging_dir)
        self.writer = tensorboard.SummaryWriter(self.logging_dir)
        logger.info(f"Initialized TensorBoard project {self.run_name} logging to {self.logging_dir}")
        logger.info("Make sure to log any initial configurations with `self.store_init_configuration` before training!")

    @property
    def tracker(self):
        return self.writer

    def store_init_configuration(self, values: dict):
        """
        Logs `values` as hyperparameters for the run. Should be run at the beginning of your experiment.

        Args:
            values (Dictionary `str` to `bool`, `str`, `float` or `int`):
                Values to be stored as initial hyperparameters as key-value pairs. The values need to have type `bool`,
                `str`, `float`, `int`, or `None`.
        """
        values = _flatten_dict(values)
        values = _sanitize_params(values)
        self.writer.add_hparams(values, metric_dict={})
        self.writer.flush()
        logger.info("Stored initial configuration hyperparameters to TensorBoard")

    def log(self, values: dict, step: Optional[int] = None):
        """
        Logs `values` to the current run.

        Args:
            values (Dictionary `str` to `str`, `float`, `int` or `dict` of `str` to `float`/`int`):
                Values to be logged as key-value pairs. The values need to have type `str`, `float`, `int` or `dict` of
                `str` to `float`/`int`.
            step (`int`, *optional*):
                The run step. If included, the log will be affiliated with this step.
        """
        for k, v in values.items():
            if isinstance(v, (int, float)):
                self.writer.add_scalar(k, v, global_step=step)
            elif isinstance(v, str):
                self.writer.add_text(k, v, global_step=step)
            elif isinstance(v, dict):
                v = _flatten_dict(v)
                self.writer.add_scalars(k, v, global_step=step)
        self.writer.flush()
        logger.info("Successfully logged to TensorBoard")

    def finish(self):
        """
        Closes `TensorBoard` writer
        """
        self.writer.close()
        logger.info("TensorBoard writer closed")


class LayerTracker(GeneralTracker):
    """
    A `Tracker` class that supports `layer`. Should be initialized at the start of your script.
    Args:
        run_name (`str`):
            The name of the experiment run.
        kwargs:
            Additional key word arguments passed along to the `layer.init` method.
    """

    name = "layer"
    requires_logging_directory = False

    def __init__(self, run_name: str, **kwargs):
        check_requirements(["layer"])

        self.run_name = run_name
        self.model_name = kwargs.get("model_name", run_name)

        api_token = kwargs.get("api_token", None)
        kwargs.pop("api_token", None)
        if api_token is not None:
            layer.login_with_access_token(api_token)
        self.project: layer.contracts.projects.Project = layer.init(project_name=self.run_name, **kwargs)
        logger.info(f"Initialized Layer project {self.run_name}")
        logger.info("Make sure to log any initial configurations with `self.store_init_configuration` before training!")

    @property
    def tracker(self):
        return layer

    def store_init_configuration(self, values: dict):
        """
        Logs `values` as hyperparameters for the run. Should be run at the beginning of your experiment.
        Args:
            values (Dictionary `str` to `bool`, `str`, `float` or `int`):
                Values to be stored as initial hyperparameters as key-value pairs. The values need to have type `bool`,
                `str`, `float`, `int`, or `None`.
        """
        layer.log({"config": values})
        logger.info("Stored initial configuration hyperparameters to Layer")

    def log(self, values: dict, step: Optional[int] = None, **kwargs):
        """
        Logs `values` to the current run.
        Args:
            values (Dictionary `str` to `str`, `float`, `int` or `dict` of `str` to `float`/`int`):
                Values to be logged as key-value pairs. The values need to have type `str`, `float`, `int` or `dict` of
                `str` to `float`/`int`.
            step (`int`, *optional*):
                The run step. If included, the log will be affiliated with this step.
            kwargs:
                Additional key word arguments passed along to the `layer.log` method.
        """
        layer.log(values, step=step, **kwargs)
        logger.info("Successfully logged to Layer")

    def finish(self):
        """
        Closes `Layer` writer
        """
        pass
        logger.info("Layer run closed")



# modified from https://github.com/facebookresearch/pytorchvideo/blob/main/pytorchvideo/data/labeled_video_paths.py
# and https://github.com/facebookresearch/pytorchvideo/blob/main/pytorchvideo/data/labeled_video_dataset.py

from __future__ import annotations

import logging
import os
import pathlib
import zipfile
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import torch
from iopath.common.file_io import g_pathmgr
from pytorchvideo.data.clip_sampling import ClipSampler
from pytorchvideo.data.labeled_video_dataset import LabeledVideoDataset as LabeledVideoDataset_
from pytorchvideo.data.labeled_video_paths import LabeledVideoPaths as LabeledVideoPaths_
from pytorchvideo.data.labeled_video_paths import make_dataset_from_video_folders
from pytorchvideo.data.video import VideoPathHandler
from torchvision.datasets.folder import make_dataset

from video_transformers.utils.file import download_file

logger = logging.getLogger(__name__)


class LabeledVideoPaths(LabeledVideoPaths_):
    """
    LabeledVideoPaths contains pairs of video path and integer index label.
    """

    labels = None

    @classmethod
    def from_path(cls, data_path: str) -> LabeledVideoPaths:
        """
        Factory function that creates a LabeledVideoPaths object depending on the path
        type.
        - If it is a directory path it uses the LabeledVideoPaths.from_directory function.
        - If it's a file it uses the LabeledVideoPaths.from_csv file.
        Args:
            file_path (str): The path to the file to be read.
        """

        if g_pathmgr.isfile(data_path):
            return LabeledVideoPaths.from_csv(data_path)
        elif g_pathmgr.isdir(data_path):
            class_0 = g_pathmgr.ls(data_path)[0]
            video_0 = g_pathmgr.ls(pathlib.Path(data_path) / class_0)[0]
            video_0_path = pathlib.Path(data_path) / class_0 / video_0
            if g_pathmgr.isfile(video_0_path):
                return LabeledVideoPaths.from_directory(data_path)
            else:
                return LabeledVideoPaths.from_directory_of_video_folders(data_path)
        else:
            raise FileNotFoundError(f"{data_path} not found.")

    @classmethod
    def from_directory(cls, dir_path: str) -> LabeledVideoPaths:
        """
        Factory function that creates a LabeledVideoPaths object by parsing the structure
        of the given directory's subdirectories into the classification labels. It
        expects the directory format to be the following:
             dir_path/<class_name>/<video_name>.mp4

        Classes are indexed from 0 to the number of classes, alphabetically.

        E.g.
            dir_path/class_x/xxx.ext
            dir_path/class_x/xxy.ext
            dir_path/class_x/xxz.ext
            dir_path/class_y/123.ext
            dir_path/class_y/nsdf3.ext
            dir_path/class_y/asd932_.ext

        Would produce two classes labeled 0 and 1 with 3 videos paths associated with each.

        Args:
            dir_path (str): Root directory to the video class directories .
        """
        assert g_pathmgr.exists(dir_path), f"{dir_path} not found."

        # Find all classes based on directory names. These classes are then sorted and indexed
        # from 0 to the number of classes.
        classes = sorted((f.name for f in pathlib.Path(dir_path).iterdir() if f.is_dir()))
        class_to_idx = {classes[i]: i for i in range(len(classes))}
        video_paths_and_label = make_dataset(dir_path, class_to_idx, extensions=("mp4", "avi"))
        assert len(video_paths_and_label) > 0, f"Failed to load dataset from {dir_path}."
        cls.labels = classes
        return cls(video_paths_and_label)

    @classmethod
    def from_directory_of_video_folders(cls, dir_path: str) -> LabeledVideoPaths:
        """
        Factory function that creates a LabeledVideoPaths object by parsing the structure
        of the given directory's subdirectories into the classification labels. It
        expects the directory format to be the following:
             dir_path/<class_name>/<video_name>/<frame_name>.jpg

        Classes are indexed from 0 to the number of classes, alphabetically.

        E.g.
            dir_path/class_x/vid1/xxx.ext
            dir_path/class_x/vid1/xxy.ext
            dir_path/class_x/vid2/xxz.ext
            dir_path/class_y/vid3/123.ext
            dir_path/class_y/vid4/nsdf3.ext
            dir_path/class_y/vid4/asd932_.ext

        Would produce two classes labeled 0 and 1 with 2 videos paths associated with each.

        Args:
            dir_path (str): Root directory to the video class directories .
        """
        assert g_pathmgr.exists(dir_path), f"{dir_path} not found."

        # Find all classes based on directory names. These classes are then sorted and indexed
        # from 0 to the number of classes.
        classes = sorted((f.name for f in pathlib.Path(dir_path).iterdir() if f.is_dir()))
        class_to_idx = {classes[i]: i for i in range(len(classes))}
        video_paths_and_label = make_dataset_from_video_folders(dir_path, class_to_idx, extensions=("jpg", "png"))
        assert len(video_paths_and_label) > 0, f"Failed to load dataset from {dir_path}."
        cls.labels = classes
        return cls(video_paths_and_label)


class LabeledVideoDataset(LabeledVideoDataset_):
    """
    LabeledVideoDataset handles the storage, loading, decoding and clip sampling for a
    video dataset. It assumes each video is stored as either an encoded video
    (e.g. mp4, avi) or a frame video (e.g. a folder of jpg, or png)
    """

    def __init__(
        self,
        labeled_video_paths: List[Tuple[str, Optional[dict]]],
        clip_sampler: ClipSampler,
        video_sampler: Type[torch.utils.data.Sampler] = torch.utils.data.RandomSampler,
        transform: Optional[Callable[[dict], Any]] = None,
        decode_audio: bool = True,
        decode_video: bool = True,
        decoder: str = "pyav",
        dataset_multiplier: int = 1,
    ):
        """
        Args:
            labeled_video_paths (List[Tuple[str, Optional[dict]]]): A list of pairs of
                video paths and optional labels.
            clip_sampler (ClipSampler): A clip sampler that samples a clip from a video.
            video_sampler (torch.utils.data.Sampler): A sampler that samples a video from the
                dataset.
            transform (Callable[[dict], Any]): A function that transforms a video dictionary
                into a tensor.
            decode_audio (bool): Whether to decode the audio of the video.
            decode_video (bool): Whether to decode the video of the video.
            decoder (str): The decoder to use for decoding the video.
            dataset_multiplier (int): The number of times to repeat the dataset.
        """

        self._decode_audio = decode_audio
        self._decode_video = decode_video
        self._transform = transform
        self._clip_sampler = clip_sampler
        self._labeled_videos = labeled_video_paths
        self._decoder = decoder

        # If a RandomSampler is used we need to pass in a custom random generator that
        # ensures all PyTorch multiprocess workers have the same random seed.
        self._video_random_generator = None
        if video_sampler == torch.utils.data.RandomSampler:
            self._video_random_generator = torch.Generator()
            self._video_sampler = video_sampler(
                self._labeled_videos,
                generator=self._video_random_generator,
                replacement=True,
                num_samples=len(self._labeled_videos) * dataset_multiplier,
            )
        else:
            self._video_sampler = video_sampler(self._labeled_videos)

        self._video_sampler_iter = None  # Initialized on first call to self.__next__()

        # Depending on the clip sampler type, we may want to sample multiple clips
        # from one video. In that case, we keep the store video, label and previous sampled
        # clip time in these variables.
        self._loaded_video_label = None
        self._loaded_clip = None
        self._last_clip_end_time = None
        self.video_path_handler = VideoPathHandler()

    @property
    def labels(self):
        """
        Returns:
            The list of class labels.
        """
        return self._labeled_videos.labels

    @property
    def videos_per_class(self):
        class_id_to_number = defaultdict(int)
        for _labeled_video in self._labeled_videos:
            label_info = _labeled_video[1]
            class_id_to_number[label_info["label"]] += 1
        class_ids = list(class_id_to_number.keys())
        return [class_id_to_number[class_id] for class_id in range(max(class_ids) + 1)]

    def __len__(self):
        if isinstance(self.video_sampler, torch.utils.data.SequentialSampler):
            return sum([1 for _ in self])
        elif isinstance(self.video_sampler, torch.utils.data.RandomSampler):
            return len(self.video_sampler)
        else:
            raise ValueError(f"Lenght calculation not implemented for sampler: {type(self.video_sampler)}.")


def labeled_video_dataset(
    data_path: str,
    clip_sampler: ClipSampler,
    video_sampler: Type[torch.utils.data.Sampler] = None,
    transform: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
    video_path_prefix: str = "",
    decode_audio: bool = True,
    decoder: str = "pyav",
    dataset_multiplier: int = 1,
) -> LabeledVideoDataset:
    """
    A helper function to create ``LabeledVideoDataset`` object for Ucf101 and Kinetics datasets.

    Args:
        data_path (str): Path to the data. The path type defines how the data
            should be read:

            * For a file path, the file is read and each line is parsed into a
              video path and label.
            * For a directory, the directory structure defines the classes
              (i.e. each subdirectory is a class).

        clip_sampler (ClipSampler): Defines how clips should be sampled from each
                video. See the clip sampling documentation for more information.

        video_sampler (Type[torch.utils.data.Sampler]): Sampler for the internal
                video container. This defines the order videos are decoded and,
                if necessary, the distributed split.

        transform (Callable): This callable is evaluated on the clip output before
                the clip is returned. It can be used for user defined preprocessing and
                augmentations to the clips. See the ``LabeledVideoDataset`` class for clip
                output format.

        video_path_prefix (str): Path to root directory with the videos that are
                loaded in ``LabeledVideoDataset``. All the video paths before loading
                are prefixed with this path.

        decode_audio (bool): If True, also decode audio from video.

        decoder (str): Defines what type of decoder used to decode a video.

    """
    labeled_video_paths = LabeledVideoPaths.from_path(data_path)
    labeled_video_paths.path_prefix = video_path_prefix
    video_sampler = torch.utils.data.RandomSampler(
        replacement=True, num_samples=len(labeled_video_paths) * dataset_multiplier
    )
    dataset = LabeledVideoDataset(
        labeled_video_paths,
        clip_sampler,
        video_sampler,
        transform,
        decode_audio=decode_audio,
        decoder=decoder,
    )
    return dataset


def download_ucf6(data_path: str):
    """
    Downloads the ucf6 dataset to the given path.
    """
    download_url = "https://github.com/fcakyon/video-transformers/releases/download/0.0.0/ucf6.zip"
    download_path = os.path.join(data_path, "ucf6.zip")
    download_file(download_url, download_path)
    with zipfile.ZipFile(download_path, "r") as zip_ref:
        zip_ref.extractall(data_path)



class TaskMixin:
    def training_step(self, batch):
        raise NotImplementedError()

    def on_training_epoch_end(self):
        raise NotImplementedError()

    def validation_step(self, batch):
        raise NotImplementedError()

    def on_validation_epoch_end(self):
        raise NotImplementedError()



from typing import Any, Tuple


def class_to_config(
    class_,
    allowed_types: Tuple[Any] = (int, float, str, dict, list, tuple, bool),
    ignored_attrs: Tuple[str] = ("config", "dump_patches", "training"),  # ignore nn.Module attributes
):
    """
    Converts a class attributes into a config dict.

    Args:
        class_: The class to convert.
        allowed_types: The attribute value types that are allowed in the config.
        ignored_attrs: The attributes that are ignored.

    Returns:
        The config dict.
    """
    config = {"name": class_.__class__.__name__}
    for attribute in dir(class_):
        if attribute not in ignored_attrs:
            if not attribute.startswith("__") and not attribute.startswith("_"):
                value = getattr(class_, attribute)
                if type(value) in allowed_types:
                    if type(value) == tuple:
                        value = list(value)
                    config[attribute] = value
    return config


def scheduler_to_config(scheduler):
    """
    Converts a scheduler and optimizer attributes into a config dict.

    Args:
        scheduler: The scheduler to convert.

    Returns:
        The config dict.
    """

    import torch

    if not isinstance(scheduler, torch.optim.lr_scheduler.SequentialLR) or not len(scheduler._schedulers) == 2:
        raise TypeError("Scheduler must be a SequentialLR with 2 sub-schedulers")

    warmup_milestone_epoch = scheduler._milestones[0]
    warmup_scheduler = scheduler._schedulers[0]
    main_scheduler = scheduler._schedulers[1]
    if isinstance(main_scheduler, torch.optim.lr_scheduler.LinearLR):
        return {
            "optimizer": {
                "name": scheduler.optimizer.__class__.__name__,
                "defaults": scheduler.optimizer.defaults,
            },
            "warmup_scheduler": {
                "class": "torch.optim.lr_scheduler.LinearLR",
                "start_factor": warmup_scheduler.start_factor,
                "end_factor": warmup_scheduler.end_factor,
                "total_iters": warmup_scheduler.total_iters,
            },
            "main_scheduler": {
                "class": "torch.optim.lr_scheduler.LinearLR",
                "start_factor": main_scheduler.start_factor,
                "end_factor": main_scheduler.end_factor,
                "total_iters": main_scheduler.total_iters,
            },
            "warmup_milestone_epoch": warmup_milestone_epoch,
        }
    elif isinstance(main_scheduler, torch.optim.lr_scheduler.CosineAnnealingLR):
        return {
            "optimizer": {
                "name": scheduler.optimizer.__class__.__name__,
                "defaults": scheduler.optimizer.defaults,
            },
            "warmup_scheduler": {
                "class": "torch.optim.lr_scheduler.LinearLR",
                "start_factor": warmup_scheduler.start_factor,
                "end_factor": warmup_scheduler.end_factor,
                "total_iters": warmup_scheduler.total_iters,
            },
            "main_scheduler": {
                "class": "torch.optim.lr_scheduler.CosineAnnealingLR",
                "T_max": main_scheduler.T_max,
            },
            "warmup_milestone_epoch": warmup_milestone_epoch,
        }
    elif isinstance(main_scheduler, torch.optim.lr_scheduler.MultiStepLR):
        return {
            "optimizer": {
                "name": scheduler.optimizer.__class__.__name__,
                "defaults": scheduler.optimizer.defaults,
            },
            "warmup_scheduler": {
                "class": "torch.optim.lr_scheduler.LinearLR",
                "start_factor": warmup_scheduler.start_factor,
                "end_factor": warmup_scheduler.end_factor,
                "total_iters": warmup_scheduler.total_iters,
            },
            "main_scheduler": {
                "class": "torch.optim.lr_scheduler.MultiStepLR",
                "milestones": main_scheduler.milestones,
                "gamma": main_scheduler.gamma,
            },
            "warmup_milestone_epoch": warmup_milestone_epoch,
        }
    else:
        raise TypeError(f"Scheduler not implemented {main_scheduler}")



import json
import os
from typing import Dict, List, Union

import torch
from huggingface_hub.constants import PYTORCH_WEIGHTS_NAME
from huggingface_hub.file_download import hf_hub_download
from huggingface_hub.hub_mixin import PyTorchModelHubMixin
from torch import nn

import video_transformers.backbones.base
from video_transformers.deployment.onnx import export
from video_transformers.heads import LinearHead
from video_transformers.utils.torch import get_num_total_params, get_num_trainable_params


class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


class TimeDistributed(nn.Module):
    """
    Time x Backbone2D (BxCxTxHxW)
           ↓
         Pool2D
           ↓
        (BxTxF)
    """

    def __init__(self, backbone: video_transformers.backbones.base.Backbone, low_memory=False):
        super(TimeDistributed, self).__init__()
        self.backbone = backbone
        self.low_memory = low_memory
        self.avg_pool = nn.AdaptiveAvgPool2d(1)

    @property
    def num_features(self):
        return self.backbone.num_features

    @property
    def mean(self):
        return self.backbone.mean

    @property
    def std(self):
        return self.backbone.std

    @property
    def type(self):
        return self.backbone.type

    @property
    def model_name(self):
        return self.backbone.model_name

    @property
    def framework(self):
        return self.backbone.framework

    @property
    def num_trainable_params(self):
        return self.backbone.num_trainable_params

    @property
    def num_total_params(self):
        return self.backbone.num_total_params

    @property
    def config(self):
        return self.backbone.config

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: BxCxTxHxW
        batch_size, num_channels, num_timesteps, height, width = x.size()

        if self.low_memory:
            output_matrix = []
            for timestep in range(num_timesteps):
                x_t = self.backbone(x[:, :, timestep, :, :])  # x_t: BxC
                output_matrix.append(x_t)
            x = torch.stack(output_matrix, dim=1)  # x: BxTxF
            # clear up memory space
            x_t = None
            output_matrix = None
        else:
            x = x.permute((0, 2, 1, 3, 4))  # x: BxTxCxHxW
            x = x.contiguous().view(batch_size * num_timesteps, num_channels, height, width)  # x: (BxT)xCxHxW
            x = self.backbone(x)  # x: (BxT)xF
            x = x.contiguous().view(batch_size, num_timesteps, x.size(1))  # x: BxTxF
        return x


class VideoClassificationModel(nn.Module, PyTorchModelHubMixin):
    """
    (BxCxTxHxW)
         ↓
      Backbone
         ↓
       Neck
         ↓
       (BxF)
         ↓
       Head

    """

    @classmethod
    def from_config(cls, config: Dict) -> "VideoClassificationModel":
        """
        Loads a model from a config file.
        """
        import video_transformers

        if isinstance(config, str):
            with open(config) as f:
                config = json.load(f)

        backbone = video_transformers.AutoBackbone.from_config(config["backbone"])
        head = video_transformers.AutoHead.from_config(config["head"])
        neck = None
        if "neck" in config:
            neck = video_transformers.AutoNeck.from_config(config["neck"])

        return cls(
            backbone=backbone,
            head=head,
            neck=neck,
            timesteps=config["num_timesteps"],
            input_size=config["preprocess_input_size"],
            labels=config["labels"],
        )

    @classmethod
    def _from_pretrained(
        cls,
        model_id,
        revision,
        cache_dir,
        force_download,
        proxies,
        resume_download,
        local_files_only,
        use_auth_token,
        map_location="cpu",
        strict=False,
        **model_kwargs,
    ):

        map_location = torch.device(map_location)

        if os.path.isdir(model_id):
            print("Loading weights from local directory")
            model_file = os.path.join(model_id, PYTORCH_WEIGHTS_NAME)
        else:
            model_file = hf_hub_download(
                repo_id=model_id,
                filename=PYTORCH_WEIGHTS_NAME,
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                use_auth_token=use_auth_token,
                local_files_only=local_files_only,
            )
        model = cls.from_config(**model_kwargs)

        state_dict = torch.load(model_file, map_location=map_location)
        model.load_state_dict(state_dict, strict=strict)
        model.eval()

        return model

    def __init__(
        self,
        backbone: Union[TimeDistributed, video_transformers.backbones.base.Backbone],
        head: LinearHead,
        neck=None,
        timesteps: int = None,
        input_size: int = None,
        labels: List[str] = None,
    ):
        """
        Args:
            backbone: Backbone model.
            head: Head model.
            neck: Neck model.
            timesteps: Number of input timesteps (required for onnx export).
            input_size: Input size of model (required for onnx export).
            labels: List of labels (required for onnx export).
        """
        super().__init__()
        self.backbone = backbone
        self.neck = neck
        self.head = head

        # required for exporting to ONNX
        self.timesteps = timesteps
        self.input_size = input_size
        self.labels = labels

    @property
    def num_features(self):
        return self.backbone.num_features

    @property
    def mean(self):
        return self.backbone.mean

    @property
    def std(self):
        return self.backbone.std

    @property
    def backbone_type(self):
        return self.backbone.type

    @property
    def backbone_name(self):
        return self.backbone.model_name

    @property
    def backbone_framework(self):
        return self.backbone.framework

    @property
    def num_trainable_params(self):
        return get_num_trainable_params(self)

    @property
    def num_total_params(self):
        return get_num_total_params(self)

    @property
    def config(self):
        config = {}
        config["backbone"] = self.backbone.config
        config["head"] = self.head.config
        if self.neck is not None:
            config["neck"] = self.neck.config
        return config

    def to_onnx(
        self,
        quantize: bool = False,
        opset_version: int = 12,
        export_dir: str = "exports/",
        export_filename: str = "model.onnx",
    ):
        """
        Export model to ONNX format.

        Args:
            quantize: Whether to quantize the model.
            opset_version: Opset version to use.
            export_dir: Directory to export model to.
            export_filename: Filename to export model to.
        """

        export(self, quantize, opset_version, export_dir, export_filename)

    @property
    def example_input_array(self):
        if self.timesteps and self.input_size:
            return torch.rand(1, 3, self.timesteps, self.input_size, self.input_size)
        else:
            return None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: BxCxTxHxW
        x = self.backbone(x)
        if self.neck is not None:
            x = self.neck(x)
        x = self.head(x)
        return x
